#include <SPI.h>
#include "ssd1306SPI.h"
#include "readFont.h"

READFONT readFont ;

// Declaration for SSD1306 display connected using software SPI (default case):
#define OLED_CS     7   // Slave Select(Chip select)
#define OLED_DC     6   // SSD1306 Data/Command
#define OLED_RESET  5   // SSD1306 Reset

static uint8_t oledAddr = 0x3C;     // SSD1306のI2Cアドレス
static uint8_t oldCommand = 0x80;   // SSD1306の1Byteコマンド
static uint8_t oldCommands = 0x00;  // SSD1306の複数Byteコマンド
static uint8_t oldData = 0x40;      // SSD1306のデータバイト

// --- SSD1306 COMMADN : I2Cで使ったのと同じ ---
static uint8_t oledInitCommands[] = { // SSD1306の初期化コマンド群
  0xC8,         // 上下の描画方向
  0xA1,         // 左右の描画方向
  0xA8, 0x3F,   // 画面の解像度
  0xD3, 0x00,   // 縦方向のオフセット
  0x40,         // 縦方向の描画開始位置
  0xA6,         // 表示方法 - A7にすると画面反転
  0x8D, 0x14    // OLEDパネルの供給電圧と思われる 14hに設定しないとダメらしい
};

// 初期化
void SSD1306SPI::init() {
  // --- SOFTWARE SPI ピン設定
  pinMode(OLED_CS, OUTPUT);   // Sleva Select
  pinMode(OLED_DC, OUTPUT);   // SSD1306 Set data/command pin as output

  //SPI初期化
  digitalWrite(OLED_CS, LOW);
  SPI.begin();//SPIを初期化、SCK、MOSI、SSの各ピンの動作は出力、SCK、MOSIはLOW、SSはHIGH
  SPI.setDataMode(SPI_MODE0);//SPIモードを「0」に設定・CPOL(クロック位相)=0,CPHA(クロック極性)=0
  SPI.setBitOrder(MSBFIRST);//SPI送受信用のビットオーダーを「MSBFIRST」に設定

  // --- Reset SSD1306
  pinMode(OLED_RESET, OUTPUT);// SSD1306 RESET
  digitalWrite(OLED_RESET, HIGH);
  delay(1);                       // VDD goes high at start, pause for 1 ms
  digitalWrite(OLED_RESET, LOW);  // Bring reset low
  delay(10);                      // Wait 10 ms
  digitalWrite(OLED_RESET, HIGH); // Bring out of reset
  delay(100);                     // Wait 10 ms

  // --- SSD1306 初期化
  // OLED設定コマンド送信
  oledSendCommands(sizeof(oledInitCommands), oledInitCommands) ;  // 初期化
  oledSendCommand(0xA4) ; // 描画ON
  oledSendCommand(0xAF) ; // 画面表示
  delay(100);                     // Wait 10 ms

  cls() ;
  dispSwitch(true) ;
}

// 画面消去
void SSD1306SPI::cls() {
  // 表示位置初期化
  cursorX = 0 ;
  cursorY = 0 ;

  // スクロールバッファ消去
  memset(DisplayBuff, 0, sizeof(DisplayBuff)) ;

  clsOLED() ;
}

void SSD1306SPI::clsOLED() {
  uint8_t SPIBuf[128] ;          // SPI転送用バッファ
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト

  for (uint8_t m = 0; m < OLED_PAGES ; m++) {
    // 表示開始位置を指定
    digitalWrite(OLED_DC, LOW); // Command mode
    SPI.transfer(0xB0 | (m & 0x07));
    SPI.transfer(0x10);
    SPI.transfer(0x00);

    // 表示データ転送
    // １バイトづつ書き込みだと折角のペリフェラルが勿体ないので
    // バッファを介して１セグメント分づつ送信している
    digitalWrite(OLED_DC, HIGH); // Data mode
    memset(SPIBuf, 0x00, 128) ;
    SPI.transfer(SPIBuf, 128);
  }
  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0);
  SPI.transfer(0x10);
  SPI.transfer(0x00);

  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// 表示日表示切替
void SSD1306SPI::dispSwitch(boolean  sw) {
  if (sw) {
    oledSendCommand(0xAF) ; // 画面表示
  } else {
    oledSendCommand(0xAE) ; // 画面表示
  }
}

// 文字列表示
void SSD1306SPI::Print(char *str, boolean  flagCR) {
  int strLen = strlen(str) ;
  uint8_t *strPos = (uint8_t *)str ;

  while (strLen > 0) {
    int utf8len = readFont.UTF8LEN((uint8_t *)strPos) ;
    uint16_t zenkaku = 0 ;
    if (utf8len <= 1) {
      uint8_t asciiCode = *strPos ;
      zenkaku = readFont.utf16_HantoZen((uint16_t)asciiCode) ;
      strPos ++ ;
      strLen -- ;
    } else {
      uint16_t utf16Code = readFont.UTF8ToUTF16(strPos) ;
      zenkaku = readFont.utf16_HantoZen(utf16Code) ;
      strPos += utf8len ;
      strLen -= utf8len ;
    }
    if (cursorX + 1 >= OLED_CHAR_W) {
      PrintCR( ) ;
    }
    DisplayBuff[cursorX + cursorY * OLED_CHAR_W] = zenkaku ;
    readFont.getFontDataByUTF16(zenkaku) ;
    oledBitMap(cursorX, cursorY, 8 , readFont.fontData) ;
    cursorX ++ ;
  }
  if (flagCR) {
    PrintCR() ;
  }
}

// 改行
void SSD1306SPI::PrintCR( ) {
  if (cursorY >= OLED_CHAR_H - 1) {
    scroll( ) ;
  } else {
    cursorY += 1 ;
  }
  cursorX = 0 ;
}

// スクロール
void SSD1306SPI::scroll() {
  uint8_t blankFont[8] ; // 空白文字用
  memset(blankFont,0,sizeof(blankFont)) ;
  memcpy(&DisplayBuff[0], &DisplayBuff[OLED_CHAR_W], sizeof(uint16_t) * OLED_CHAR_W * (OLED_CHAR_H-1)) ;
  memset(&DisplayBuff[OLED_CHAR_W * (OLED_CHAR_H -1)], 0, sizeof(uint16_t) * OLED_CHAR_W) ;

  for (int y = 0 ; y < OLED_CHAR_H ; y ++ ) {
    for (int x = 0 ; x < OLED_CHAR_W ; x ++) {
      uint16_t utf16code = DisplayBuff[y * OLED_CHAR_W + x] ;
      if (utf16code == 0) {
        oledBitMap(x, y , 8 , blankFont) ;
      } else {
        readFont.getFontDataByUTF16(utf16code) ;
        oledBitMap(x, y , 8 , readFont.fontData) ;
      }
    }
  }
}

// OLEDに1バイトコマンドを送信
void SSD1306SPI::oledSendCommand(uint8_t command) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに2バイトコマンドを送信
void SSD1306SPI::oledSendCommand2(uint8_t command1, uint8_t command2) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command1);
  SPI.transfer(command2);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに3バイトコマンドを送信
void SSD1306SPI::oledSendCommand3(uint8_t command1, uint8_t command2, uint8_t command3) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(command1);
  SPI.transfer(command2);
  SPI.transfer(command3);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// OLEDに複数のコマンドを送信
void SSD1306SPI::oledSendCommands(int commandLen, uint8_t commands[]) {
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(commands, commandLen);
  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}

// 128x32ドットＯＬＥＤにビットマップを書き込み
void SSD1306SPI::oledBitMap(int col , int row , int width , uint8_t *bitMap) {
  int offsetX = col * 8 ;
  digitalWrite(OLED_CS, LOW); // SSD1306セレクト

  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0 | (row & 0x07));
  SPI.transfer(0x10 | (offsetX & 0xf0) >> 4 );
  SPI.transfer(offsetX & 0x0F);

  // 表示データ転送
  // Arduinoだとメモリが足りなくて１画面分を一気に転送する事ができない
  // なので、１セグメント分づつ送信している
  digitalWrite(OLED_DC, HIGH); // Data mode
  SPI.transfer(bitMap, width);

  // 表示開始位置を指定
  digitalWrite(OLED_DC, LOW); // Command mode
  SPI.transfer(0xB0);
  SPI.transfer(0x10);
  SPI.transfer(0x00);

  digitalWrite(OLED_CS, HIGH);  // SSD1306解放
}
